import javax.microedition.lcdui.*;
import javax.microedition.m3g.*;
import java.util.*;


/** Classe de gestion du jeu et de l'interface graphique.
 * @author Batrice Frey et Cyril Crassin
 * */
public class TicTacCanvas extends Canvas {

	/**Contexte de rendu Mobile 3D.*/
	private Graphics3D  graphics3d;
	/**Camera de la scene.*/
	private Camera   camera;
	/**Source de lumire.*/
	private Light   light;
	
	/**Angle de rotation pour l'annimation de la camera*/
	private float   angle = 0.0f;
	/**Angle de l'icone du joueur.*/
	private float   angleIcon = 0.0f;
	
	/**Definition du fond*/
	private Background  background = new Background();
	
	/**Maillages des lments du plateau.*/
 	private Mesh mTorus=null;
	private Mesh mCross=null;
	private Mesh mPlateau=null;
	private Mesh mSelect=null;
	
	/**Modes de composition des pions.*/
	CompositingMode compModeAlpha=new CompositingMode();
	CompositingMode compModeOpaque=new CompositingMode();
	
	/**Transformations.*/
	Transform camTransform = new Transform();
	Transform viewTransform = new Transform();
	Transform caseTransform = new Transform();
	Transform tmpTransform = new Transform();
	
	
	/**Plateau de jeu.*/
	private int plateau[][][];
	
	/**Positionnement du curseur en Z.*/
	private int curZ=0;
	private int curZTrans=0;
	
	/**Positionnement du curseur en X.*/
	private int curX=0;
	private int curXTrans=0;
	
	/**Positionnement du curseur en Y.*/
	private int curY=0;
	private int curYTrans=0;
	

	/**Numero du joueur.*/
	private int playerNum=0;
	
	/**Module de communication asynchrone.*/
	TicTacCom com;
	
	/**Application source.*/
	TicTacMain main;
	
	/**En attente de l'autre joueur.*/
	boolean waitToPlay=false;
	
	/**Le joueur a gagn.*/
	boolean victory=false;
	/**Le joueur a perdu.*/
	boolean defeat=false;
	/**Match nul*/
	boolean nullMatch=false;
	/**Nombre de coups dja jous*/
	private int nbCoups=0;
	/**Nombre de coups maximum avant match nul*/
	private int maxCoups=27;
	
	
	/**
	 * Construction du jeu.
	 * @param m Application source.
	 * @param pn Numero du joueur.
	 * @param c Module de communication.
	 * */
	public TicTacCanvas(TicTacMain m, int pn, TicTacCom c) {
	 	playerNum=pn;
	 	com=c;
	 	main=m;
	 	
		// set up this Displayable to listen to command events
		setCommandListener(new CommandListener() {
			public void commandAction(Command c, Displayable d) {
				if (c.getCommandType() == Command.EXIT) 
					TicTacMain.quitApp();
				else if(c.getCommandType() == Command.ITEM)
					main.restartApp();
				
			}
			
		  	});
		
		try { init();}
		catch(Exception e) { e.printStackTrace();}
	 }

	 
	/**
	 * Initialisation du jeu.
	 * Mise en place des lments 3D.
	 * Debut du jeu.
	 * @throws Exception
	 */
	 private void init() throws Exception {
	 	
	 	////Canvas
	  addCommand(new Command("Exit", Command.EXIT, 1));
	  addCommand(new Command("Restart", Command.ITEM, 1));
	  
	  graphics3d = Graphics3D.getInstance();
	
	  camera = new Camera();
	  camera.setPerspective( 60.0f,
	   (float)getWidth()/ (float)getHeight(),
	   1.0f,
	   1000.0f );
	
	  light = new Light();
	  light.setColor(0xffffff);
	  light.setIntensity(1.25f);
	
	
	  
	  //background.setColor(0xffffcc);
	  background.setColor(0xDDEEDD);
	  
	  Object3D roots[]=Loader.load("/TicTac7.m3g");//4 beau
	  Group tGroup=(Group)roots[0];
	  
	  for(int i=0; i<tGroup.getChildCount(); i++)
	  	if(tGroup.getChild(i).getUserID()==100){
	  		mTorus=(Mesh)tGroup.getChild(i);
	  	}else if(tGroup.getChild(i).getUserID()==101){
	  		mCross=(Mesh)tGroup.getChild(i);
	  	}else if(tGroup.getChild(i).getUserID()==102){
		  	mPlateau=(Mesh)tGroup.getChild(i);
	  	}else if(tGroup.getChild(i).getUserID()==103){
		  	mSelect=(Mesh)tGroup.getChild(i);
	  	}
	  	
	  mTorus.setAlphaFactor(0.3f);
	  mCross.setAlphaFactor(0.3f);
	  compModeAlpha.setBlending(CompositingMode.ALPHA);
	  //compModeAlpha.setBlending(CompositingMode.REPLACE );
	  compModeOpaque.setBlending(CompositingMode.REPLACE);
	  
	  
	  plateau=new int[3][3][3];
	  
	  
	  if(playerNum==1)
	  	waitToPlay=false;
	  else{
	  	waitToPlay=true;
	  	com.receive();
	  }
	  
	  //Camera setup
	  camTransform.postTranslate(0.0f, 0.0f, 30.0f);
	  graphics3d.setCamera(camera, camTransform);
	  
	  // Set light
	  graphics3d.resetLights();
	  graphics3d.addLight(light, camTransform);
	  
	 }

	 /**
	  * Dessin d'un pion tore.
	  * @param transform Position du pion.
	  */
	 protected void drawTorus(Transform transform){
	 	tmpTransform.set(transform);
	 	tmpTransform.postRotate(90.0f, 1.0f, 0.0f, 0.0f);
		graphics3d.render(mTorus, tmpTransform);
	 }
	 
	 /**
	  * Dessin d'un pion croix.
	  * @param transform Position du pion.
	  */
	 protected void drawCross(Transform transform){
	 	tmpTransform.set(transform);
	 	tmpTransform.postRotate(45.0f, 0.0f, 1.0f, 0.0f);
		graphics3d.render(mCross, tmpTransform);
	 }
	 
	 /**
	  * Dessin de l'outils de selection.
	  * @param transform Position de la selection.
	  */
	 protected void drawSelect(Transform transform){
	 	tmpTransform.set(transform);
	 	tmpTransform.postScale(0.9f, 0.9f, 0.9f);
		graphics3d.render(mSelect, tmpTransform);
	 }
	 
	 /**
	  * Rafraichissement de la scene.
	  * Appele a interval regulier par le timer de TicTacMain.
	  * Dessine l'ensemble de la scene.
	  * @param g Zone de dessin.
	  */
	 protected void paint(Graphics g) {
		 graphics3d.bindTarget(g, true,
		 		Graphics3D.DITHER |
				Graphics3D.TRUE_COLOR);
		 graphics3d.clear(background);
		
		 
		 if(waitToPlay && com.newReceivedData){
		 		int []t=com.getReceivedData();
		 		plateau[t[1]][t[2]][t[3]]=t[0];
		 		nbCoups++;
		 		
		 		
		 		waitToPlay=false;
		 		
		 		if(gagne(t[0], t[1], t[2], t[3])){
					defeat=true;
					waitToPlay=true;
		 		}else if(nbCoups==maxCoups)
		 			nullMatch=true;
		 }
		 	
		 
		 if(curZTrans < curZ*10)
		 	curZTrans+=1;
		 else if(curZTrans > curZ*10)
		 	curZTrans-=1;
		 
		 if(curXTrans < curX*10)
		 	curXTrans+=1;
		 else if(curXTrans > curX*10)
		 	curXTrans-=1;
		 
		
		 //Icone du joueur
		 
		 angleIcon+=5.0f;
		 
		 tmpTransform.setIdentity();
		 tmpTransform.postTranslate(11.0f, 15.0f, -0.01f);
		 tmpTransform.postScale(0.1f, 0.1f, 0.1f);
		 if(!waitToPlay)
		 	tmpTransform.postRotate(angleIcon, 1.0f, 1.0f, 1.0f );
		 else
		 	tmpTransform.postScale(0.5f+(float)Math.sin(angleIcon/100.0f)/2.0f, 
		 				0.5f+(float)Math.sin(angleIcon/100.0f)/2.0f, 
		 				0.5f+(float)Math.sin(angleIcon/100.0f)/2.0f );
		 
		 if( playerNum==1){
		 	mTorus.getAppearance(0).setCompositingMode(compModeOpaque);
			drawTorus(tmpTransform);
		 }else if( playerNum==2){
		 	mCross.getAppearance(0).setCompositingMode(compModeOpaque);
			drawCross(tmpTransform);
		 }
			
			
			
		// rotation de la scene
		angle += 1.0f;
		viewTransform.setIdentity();
		viewTransform.postTranslate(0.0f, 0.0f, -200.0f);
		viewTransform.postRotate((curZTrans*1.80f)+10.0f, 1.0f, 0.0f, 0.0f);
		viewTransform.postRotate(( (20-curXTrans)*2.0f)-20.0f, 0.0f, 1.0f, 0.0f);
		 
		
		  for(int j=0; j<3; j++)
		  	for(int k=0; k<3; k++)
				 for(int i=0; i<3; i++){
				 	caseTransform.set(viewTransform);
				 	caseTransform.postTranslate( (i-1)*60.0f, (k-1)*50.0f, (j-1)*60.0f );
		  			if(k==curZ){
		  				caseTransform.postScale(1.3f, 1.3f, 1.3f);
		  				mTorus.getAppearance(0).setCompositingMode(compModeOpaque);
		  				mCross.getAppearance(0).setCompositingMode(compModeOpaque);
		  			}else{
		  				caseTransform.postScale(1.2f, 1.2f, 1.2f);
		  				mTorus.getAppearance(0).setCompositingMode(compModeAlpha);
		  				mCross.getAppearance(0).setCompositingMode(compModeAlpha);
		  			}
		  			
		  			if( plateau[i][j][k]==1)
		  				drawTorus(caseTransform);
		  			else if( plateau[i][j][k]==2)
		  				drawCross(caseTransform);
		  			
		  			if(i==curX && j==curY && k==curZ)
		  				drawSelect(caseTransform);
		 
		  		}

		  

		graphics3d.render(mPlateau, viewTransform);
		graphics3d.releaseTarget();
		
		
		if(waitToPlay && !com.newReceivedData){
			g.setColor(0xFF6611);
			g.setFont( Font.getFont( Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE ));
			
			g.drawString("L'autre joueur joue...", getWidth()/2, getHeight()/2 - 10, 
		              Graphics.HCENTER | Graphics.BASELINE);
		}else if(victory){
			g.setColor(0xFF1111);
			g.setFont( Font.getFont( Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE ));
			
			g.drawString("Victoire !!! :-)", getWidth()/2, getHeight()/2 - 10, 
		              Graphics.HCENTER | Graphics.BASELINE);
		}else if(defeat){
			g.setColor(0xFF1111);
			g.setFont( Font.getFont( Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE ));
			
			g.drawString("Perdu ! :-(", getWidth()/2, getHeight()/2 - 10, 
		              Graphics.HCENTER | Graphics.BASELINE);
		}else if(nullMatch){
			g.setColor(0xFF1111);
			g.setFont( Font.getFont( Font.FACE_SYSTEM, Font.STYLE_BOLD, Font.SIZE_LARGE ));
			
			g.drawString("Match nul !", getWidth()/2, getHeight()/2 - 10, 
		              Graphics.HCENTER | Graphics.BASELINE);
		}
		
	 }
 
	/**
	 * Callback d'appui sur une touche.
	 * @param keyCode Code de la touche.
	 */
	protected void keyPressed(int keyCode) {
		if(!waitToPlay && !victory && !defeat && !nullMatch)
		switch(this.getGameAction(keyCode)) {
			
			//Level
			case Canvas.GAME_C  :
				if(curZ<2)
					curZ++;
				break;
			case Canvas.GAME_A :
				if(curZ>0)
					curZ--;
				break;
			//Horizontal
			case Canvas.RIGHT  :
				if(curX<2)
					curX++;
				break;
			case Canvas.LEFT :
				if(curX>0)
					curX--;
				break;
				
			//Vertical
			case Canvas.UP  :
				if(curY<2)
					curY++;
				break;
			case Canvas.DOWN :
				if(curY>0)
					curY--;
				break;
				
			case Canvas.FIRE:
				if(plateau[curX][curY][curZ]==0){
					plateau[curX][curY][curZ]=playerNum;
					nbCoups++;
					
					repaint();
					
					com.send(playerNum, curX, curY, curZ);
					
					if(gagne(playerNum, curX, curY, curZ))
						victory=true;
					else if(nbCoups==maxCoups)
			 			nullMatch=true;
					else
						com.receive();
					
					waitToPlay=true;
				}

				break;
			//Canvas.KEY_NUM7 
		}
			
	}
	
	
	/**
	 * Recuperation du contenu d'une case.
	 * @param x position en x de la case.
	 * @param y position en y de la case.
	 * @param z position en z de la case.
	 * @return La valeur de la case.
	 */
	int getCase(int x, int y, int z){
		if(x<0 || y<0 || z<0 || x>2 || y>2 || z>2)
			return 0;
		else 
			return plateau[x][y][z];
	}
	
	/**
	 * Detection de la victoire d'un joueur apres un coup.
	 * @param p Joueur a tester.
	 * @param x position en x de la case.
	 * @param y position en y de la case.
	 * @param z position en z de la case.
	 * @return Si le joueur a gagn.
	 */
	boolean gagne(int p, int x, int y, int z){
		for(int i=-1; i<2; i++)
			for(int j=-1; j<2; j++)
				for(int k=-1; k<2; k++)
					if( !(i==0 && j==0 && k==0) ){
						if( (getCase(x+i, y+j, z+k)==p) && ( (getCase(x+i*2, y+j*2, z+k*2)==p) || (getCase(x-i, y-j, z-k)==p) ) ){
							return true;
						}
					}
					
		
		return false;
					
	}
}